""" opentrons.system: Functions and data for interacting with the OT2 system.

The functions in the modules contained in this package do things like bridge
Python calls with shell invocations of nmcli or other system utilities. They
are in general unsafe to call when not running on an Opentrons robot.
"""
